
/*  GNU SED, a batch stream editor.
    Copyright (C) 1989, 1990, 1991 Free Software Foundation, Inc.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

#include "stdafx.h"
#include "RegExp.h"
#include "Engine.h"

#include "sed.h"

/* Various error messages we may want to print */
static char ONE_ADDR[]="Command only uses one address";
static char NO_ADDR[]="Command doesn't take any addresses";
static char LINE_JUNK[]="Extra characters after command";
static char BAD_EOF[]="Unexpected End-of-file";
static char NO_REGEX[]="No previous regular expression";
static char NO_COMMAND[]="Missing command";

extern int globalerror;

// *************************************************** Hilfsfunktionen

#define INVERTESCAPE "(){}?+"
long Engine::regexpcpy(char *tmp1, char *regexp)
{
   if (!regexp) return 0;
   LPSTR tmp2=(LPSTR)regexp;
   long len=strlen(tmp2), z=0;
   while (len)
   {
      if (*tmp2=='\\' && len>1 && strchr(INVERTESCAPE, *(tmp2+1))) { tmp2++; len--; }
      else if (*tmp2=='/' || strchr(INVERTESCAPE, *tmp2)) { z++; if (tmp1) *(tmp1++)='\\'; }
      if (tmp1) *(tmp1++)=*tmp2; 
      z++; tmp2++; len--;
   }
   return z;
}

void Engine::scan_regexp(char *regexp, char *expr, char cmd, int pos)
{
   char *e_strings=NULL;
   struct sed_label *go, *lbl;

   // strcpy(e_strings, regexp);
   long anzahl=14; // 7 Zeichen und "g" oder Zahl am Ende
   anzahl+=regexpcpy(0, regexp);
   anzahl+=regexpcpy(0, expr);

   LPSTR tmp1=e_strings=(char*)ck_malloc(anzahl);

   *(tmp1++)=cmd; *(tmp1++)='/'; tmp1+=regexpcpy(tmp1, regexp);
   *(tmp1++)='/';                tmp1+=regexpcpy(tmp1, expr);
   *(tmp1++)='/'; 
   if (!pos) *(tmp1++)='g'; else tmp1+=sprintf(tmp1, "%d", pos);
   *(tmp1++)='\n'; *(tmp1++)=0; *(tmp1++)=0;

   compile_string(e_strings);
   free(e_strings);

   for (go=jumps; go; go=go->next)
   {
      for (lbl=labels; lbl; lbl=lbl->next)
         if (!strcmp (lbl->name, go->name)) break;
      if (*go->name && !lbl) panic("Can't find label for jump to '%s'", go->name);
      go->v->v[go->v_index].x.jump=lbl;
   }

   line.length=0;
   line.alloc=50;
   line.text=(char*)ck_malloc (50);

   append.length=0;
   append.alloc=50;
   append.text=(char*)ck_malloc (50);

   hold.length=1;
   hold.alloc=50;
   hold.text=(char*)ck_malloc (50);
   hold.text[0]='\n';
}

/*
void mycopy(char **dst, BSTR *src)
{
#ifdef WIN32
   *dst=(LPSTR)(*src);
#else
   int len=SysStringLen(*src); 
   *dst=ck_malloc(len+2); 
   strncpy(*dst, (LPSTR)(*src), len); 
   (*dst)[len]=0; (*dst)[len+1]=0;
#endif
}

void myfree(char *dst)
{
#ifndef WIN32
   free (dst);
#endif
}
*/

// *************************************************** 

/* 'str' is a string (from the command line) that contains a sed command.
   Compile the command, and add it to the end of 'the_program' */
void Engine::compile_string (char *str)
{
   prog_file=0;
   prog_line=0;
   prog_start=prog_cur=(unsigned char *)str;
   prog_end=(unsigned char *)str+strlen (str);
   the_program=compile_program(the_program, prog_line);
}

/* 'str' is the name of a file containing sed commands.  Read them in
   and add them to the end of 'the_program' */
void Engine::compile_file (char *str)
{
   int ch;
   prog_start=prog_cur=prog_end=0;
   prog_name=str;
   prog_line=1;
   // if (str[0] == '-' && str[1] == '\0') prog_file=stdin; else
   prog_file=ck_fopen (str, "r");
   ch=getc (prog_file);
   if (ch == '#')
   {
      ch=getc (prog_file);
      if (ch == 'n') no_default_output++;
      while (ch != EOF && ch != '\n')
      {
         ch=getc (prog_file);
         if (ch == '\\') ch=getc (prog_file);
      }
      ++prog_line;
   }
   else if (ch != EOF)
      ungetc (ch, prog_file);
   the_program=compile_program (the_program, prog_line);
}

#define MORE_CMDS 40

/* Read a program (or a subprogram within '{' '}' pairs) in and store
   the compiled form in *'vector'  Return a pointer to the new vector.  */
vector *Engine::compile_program (struct vector *vector, int open_line)
{
   struct sed_cmd *cur_cmd;
   int ch=0;
   int pch;
   int slash;
   VOID *b;
   unsigned char *string;
   int num;

   if (!vector)
   {
      vector=(struct vector *) ck_malloc (sizeof (struct vector));
      vector->v=(struct sed_cmd *) ck_malloc (MORE_CMDS * sizeof (struct sed_cmd));
      vector->v_allocated=MORE_CMDS;
      vector->v_length=0;
      vector->return_v=0;
      vector->return_i=0;
   }
   for (;;)
   {
      skip_comment:
      do
      {
         pch=ch;
         ch=inchar ();
         if ((pch == '\\') && (ch == '\n')) ch=inchar ();
      }
      while (ch != EOF && (isblank (ch) || ch == '\n' || ch == ';'));
      if (ch == EOF) break;
      savchar (ch);
      if (vector->v_length == vector->v_allocated)
      {
         vector->v=((struct sed_cmd *)
         ck_realloc ((VOID *) vector->v, ((vector->v_length + MORE_CMDS) * sizeof (struct sed_cmd))));
         vector->v_allocated += MORE_CMDS;
      }
      cur_cmd = vector->v + vector->v_length;
      vector->v_length++;

      cur_cmd->a1.addr_type = 0;
      cur_cmd->a2.addr_type = 0;
      cur_cmd->aflags = 0;
      cur_cmd->cmd = 0;

      if (compile_address (&(cur_cmd->a1)))
      {
         ch = inchar ();
         if (ch == ',')
       {
         do
      ch = inchar ();
         while (ch != EOF && isblank (ch));
         savchar (ch);
         if (compile_address (&(cur_cmd->a2)))
      ;
         else
      bad_prog ("Unexpected ','");
       }
     else
       savchar (ch);
   }
      if (cur_cmd->a1.addr_type == addr_is_num
     && cur_cmd->a2.addr_type == addr_is_num
     && cur_cmd->a2.addr_number < cur_cmd->a1.addr_number)
   cur_cmd->a2.addr_number = cur_cmd->a1.addr_number;

      ch = inchar ();
      if (ch == EOF)
   bad_prog (NO_COMMAND);
    new_cmd:
      switch (ch)
   {
   case '#':
     if (cur_cmd->a1.addr_type != 0)
       bad_prog (NO_ADDR);
     do
       {
         ch = inchar ();
         if (ch == '\\')
      ch = inchar ();
       }
     while (ch != EOF && ch != '\n');
     vector->v_length--;
     goto skip_comment;
   case '!':
     if (cur_cmd->aflags & ADDR_BANG_BIT)
       bad_prog ("Multiple '!'s");
     cur_cmd->aflags |= ADDR_BANG_BIT;
     do
       ch = inchar ();
     while (ch != EOF && isblank (ch));
     if (ch == EOF)
       bad_prog (NO_COMMAND);
#if 0
     savchar (ch);
#endif
     goto new_cmd;
   case 'a':
   case 'i':
     if (cur_cmd->a2.addr_type != 0)
       bad_prog (ONE_ADDR);
     /* Fall Through */
   case 'c':
     cur_cmd->cmd = ch;
     if (inchar () != '\\' || inchar () != '\n')
       bad_prog (LINE_JUNK);
     b = init_buffer ();
     while ((ch = inchar ()) != EOF && ch != '\n')
       {
         if (ch == '\\')
      ch = inchar ();
         add1_buffer (b, ch);
       }
     if (ch != EOF)
       add1_buffer (b, ch);
     num = size_buffer (b);
     string = (unsigned char *) ck_malloc (num);
     bcopy (get_buffer (b), string, num);
     flush_buffer (b);
     cur_cmd->x.cmd_txt.text_len = num;
     cur_cmd->x.cmd_txt.text = (char *) string;
     break;
   case '{':
     cur_cmd->cmd = ch;
     program_depth++;
     cur_cmd->x.sub = compile_program ((struct vector *) 0, prog_line);
     /* FOO JF is this the right thing to do?
            almost.  don't forget a return addr.  -t */
     cur_cmd->x.sub->return_v = vector;
     cur_cmd->x.sub->return_i = vector->v_length - 1;
     break;
   case '}':
     if (!program_depth)
       bad_prog ("Unexpected '}'");
     --program_depth;
     /* a return insn for subprograms -t */
     cur_cmd->cmd = ch;
     if (cur_cmd->a1.addr_type != 0)
       bad_prog ("} doesn't want any addresses");
     while ((ch = inchar ()) != EOF && ch != '\n' && ch != ';')
       if (!isblank (ch))
         bad_prog (LINE_JUNK);
     return vector;
   case ':':
     cur_cmd->cmd = ch;
     if (cur_cmd->a1.addr_type != 0)
       bad_prog (": doesn't want any addresses");
     labels = setup_jump (labels, cur_cmd, vector);
     break;
   case 'b':
   case 't':
     cur_cmd->cmd = ch;
     jumps = setup_jump (jumps, cur_cmd, vector);
     break;
   case 'q':
   case '=':
     if (cur_cmd->a2.addr_type)
       bad_prog (ONE_ADDR);
     /* Fall Through */
   case 'd':
   case 'D':
   case 'g':
   case 'G':
   case 'h':
   case 'H':
   case 'l':
   case 'n':
   case 'N':
   case 'p':
   case 'P':
   case 'x':
     cur_cmd->cmd = ch;
     do
       ch = inchar ();
     while (ch != EOF && isblank (ch) && ch != '\n' && ch != ';');
     if (ch != '\n' && ch != ';' && ch != EOF)
       bad_prog (LINE_JUNK);
     break;

   case 'r':
     if (cur_cmd->a2.addr_type != 0)
       bad_prog (ONE_ADDR);
     /* FALL THROUGH */
   case 'w':
     cur_cmd->cmd = ch;
     cur_cmd->x.io_file = compile_filename (ch == 'r');
     break;

   case 's':
   case 'z':
   case 'f':
     cur_cmd->cmd = ch;
     slash = inchar ();
     buffer_regex (slash);
     compile_regex ();

     cur_cmd->x.cmd_regex.regx = last_regex;

     b=init_buffer ();
     while ((ch=inchar())!=EOF && ch!=slash && ch!='\n')
     {
        if (ch=='\\')
        {
           int ci=inchar();
           if (ci!=EOF)
           {
              if (ci != '\n') add1_buffer (b, ch);
              add1_buffer (b, ci);
           }
        }
        else
           add1_buffer (b, ch);
     }
     if (ch != slash)
     {
        if (ch == '\n' && prog_line > 1)
        --prog_line;
        bad_prog ("Unterminated `s' command");
     }
     cur_cmd->x.cmd_regex.replace_length = size_buffer (b);
     cur_cmd->x.cmd_regex.replacement = (char*)ck_malloc (cur_cmd->x.cmd_regex.replace_length);
     bcopy (get_buffer (b), cur_cmd->x.cmd_regex.replacement, cur_cmd->x.cmd_regex.replace_length);
     flush_buffer (b);

     cur_cmd->x.cmd_regex.flags = 0;
     cur_cmd->x.cmd_regex.numb = 0;

     if (ch==EOF) break;
     do
     {
        ch=inchar();
        switch (ch)
        {
        case 'p':
          cur_cmd->x.cmd_regex.flags |= S_PRINT_BIT;
          break;
        case 'g':
          if (cur_cmd->x.cmd_regex.flags & S_NUM_BIT) cur_cmd->x.cmd_regex.flags &= ~S_NUM_BIT;
          cur_cmd->x.cmd_regex.flags |= S_GLOBAL_BIT;
          break;
        case 'w':
          cur_cmd->x.cmd_regex.flags |= S_WRITE_BIT;
          cur_cmd->x.cmd_regex.wio_file = compile_filename(0);
          ch='\n';
          break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
          if ((cur_cmd->x.cmd_regex.flags & S_GLOBAL_BIT) == 0)
             cur_cmd->x.cmd_regex.flags |= S_NUM_BIT;
          num = 0;
          while (isdigit(ch)) { num=num*10+ch-'0'; ch=inchar(); }
          savchar(ch);
          cur_cmd->x.cmd_regex.numb = num;
          break;
        }
     }
     while (ch!=EOF && ch!='\n' && ch!=';');
     if (ch==EOF) break;
     break;

   case 'y':
     cur_cmd->cmd = ch;
     string = (unsigned char *) ck_malloc (256);
     for (num = 0; num < 256; num++)
       string[num] = num;
     b = init_buffer ();
     slash = inchar ();
     while ((ch = inchar ()) != EOF && ch != slash)
       add1_buffer (b, ch);
     cur_cmd->x.translate = string;
     string = (unsigned char *) get_buffer (b);
     for (num = size_buffer (b); num; --num)
       {
         ch = inchar ();
         if (ch == EOF)
      bad_prog (BAD_EOF);
         if (ch == slash)
      bad_prog ("strings for y command are different lengths");
         cur_cmd->x.translate[*string++] = ch;
       }
     flush_buffer (b);
     if (inchar () != slash || ((ch = inchar ()) != EOF && ch != '\n' && ch != ';'))
       bad_prog (LINE_JUNK);
     break;

   default:
     bad_prog ("Unknown command");
   }
   }
   if (program_depth)
   {
      prog_line=open_line;
      bad_prog("Unmatched `{'");
   }
   return vector;
}

/* Complain about a programming error and exit. */
void bad_prog (char *why)
{
   globalerror=TRUE;
   /*
   if (prog_line > 0)
      fprintf (stderr, "file %s line %d: %s\n", prog_name, prog_line, why);
   else
      fprintf (stderr, "(1) %s\n", why);
   */
}

/* Read the next character from the program.  Return EOF if there isn't
   anything to read.  Keep prog_line up to date, so error messages can
   be meaningful. */
int Engine::inchar (void)
{
   int ch;
   if (prog_file)
   {
      if (feof (prog_file)) return EOF;
      ch = getc (prog_file);
   }
   else
   {
      if (!prog_cur) return EOF;
      if (prog_cur == prog_end) { ch = EOF; prog_cur = 0; } else ch = *prog_cur++;
   }
   if ((ch == '\n') && prog_line) prog_line++;
   return ch;
}

/* unget 'ch' so the next call to inchar will return it.  'ch' must not be
   EOF or anything nasty like that. */
void Engine::savchar (int ch)
{
   if (ch == EOF) return;
   if (ch == '\n' && prog_line > 1) --prog_line;
   if (prog_file) ungetc (ch, prog_file); else *--prog_cur = ch;
}


/* Try to read an address for a sed command.  If it succeeeds,
   return non-zero and store the resulting address in *'addr'.
   If the input doesn't look like an address read nothing
   and return zero. */
int Engine::compile_address (struct addr *addr)
{
  int ch;
  int num;

  ch = inchar ();

  if (isdigit (ch))
    {
      num = ch - '0';
      while ((ch = inchar ()) != EOF && isdigit (ch))
   num = num * 10 + ch - '0';
      if (ch == '~')
   {
     addr->addr_type = addr_is_mod;
     addr->offset = num;
     ch = inchar();
     num=0;
     if (isdigit(ch)) {
       num = ch - '0';
       while ((ch = inchar ()) != EOF && isdigit (ch))
         num = num * 10 + ch - '0';
       addr->modulo = num;
     }
     addr->modulo += (addr->modulo==0);
   }
      else
   {
     addr->addr_type = addr_is_num;
     addr->addr_number = num;
   }
      while (ch != EOF && isblank (ch))
   ch = inchar ();
      savchar (ch);
      return 1;
    }
  else if (ch == '/' || ch == '\\')
    {
      addr->addr_type = addr_is_regex;
      if (ch == '\\')
   ch = inchar ();
      buffer_regex (ch);
      compile_regex ();
      addr->addr_regex = last_regex;
      do
   ch = inchar ();
      while (ch != EOF && isblank (ch));
      savchar (ch);
      return 1;
    }
  else if (ch == '$')
    {
      addr->addr_type = addr_is_last;
      do
   ch = inchar ();
      while (ch != EOF && isblank (ch));
      savchar (ch);
      return 1;
    }
  else
    savchar (ch);
  return 0;
}

void Engine::buffer_regex (int slash)
{
  VOID *b;
  int ch;
  int char_class_pos = -1;

  b = init_buffer ();
  while ((ch = inchar ()) != EOF && (ch != slash || (char_class_pos >= 0)))
    {
      if (ch == '^')
   {
     if (size_buffer (b) == 0)
       {
         add1_buffer (b, '\\');
         add1_buffer (b, '`');
       }
     else
       add1_buffer (b, ch);
     continue;
   }
      else if (ch == '$')
   {
     ch = inchar ();
     savchar (ch);
     if (ch == slash)
       {
         add1_buffer (b, '\\');
         add1_buffer (b, '\'');
       }
     else
       add1_buffer (b, '$');
     continue;
   }
      else if (ch == '[')
   {
     if (char_class_pos < 0)
       char_class_pos = size_buffer (b);
     add1_buffer (b, ch);
     continue;
   }
      else if (ch == ']')
   {
     add1_buffer (b, ch);
     {
       char * regexp = get_buffer (b);
       int pos = size_buffer (b) - 1;
       if (!(   (char_class_pos >= 0)
        && (   (pos == char_class_pos + 1)
            || (   (pos == char_class_pos + 2)
           && (regexp[char_class_pos + 1] == '^')))))
         char_class_pos = -1;
       continue;
     }
   }
      else if (ch != '\\' || (char_class_pos >= 0))
   {
     add1_buffer (b, ch);
     continue;
   }
      ch = inchar ();
      switch (ch)
   {
   case 'n':
     add1_buffer (b, '\n');
     break;
#if 0
   case 'b':
     add1_buffer (b, '\b');
     break;
   case 'f':
     add1_buffer (b, '\f');
     break;
   case 'r':
     add1_buffer (b, '\r');
     break;
   case 't':
     add1_buffer (b, '\t');
     break;
#endif /* 0 */
   case EOF:
     break;
   default:
     add1_buffer (b, '\\');
     add1_buffer (b, ch);
     break;
   }
    }
  if (ch == EOF)
    bad_prog (BAD_EOF);
  if (size_buffer (b))
    {
      if (last_regex_string)
   last_regex_string = (char *)ck_realloc (last_regex_string,
                  size_buffer (b) + 1);
      else
   last_regex_string = (char *)ck_malloc (size_buffer (b) + 1);
      bcopy (get_buffer (b), last_regex_string, size_buffer (b));
      last_regex_string [size_buffer (b)] = 0;
    }
  else if (!last_regex)
    bad_prog (NO_REGEX);
  flush_buffer (b);
}

void Engine::compile_regex (void)
{
  const char * error;
  last_regex = ((struct re_pattern_buffer *)ck_malloc (sizeof (struct re_pattern_buffer)));
  bzero (last_regex, sizeof (*last_regex));
  last_regex->fastmap = (char*)ck_malloc (256);
  error = re_compile_pattern (last_regex_string, strlen (last_regex_string), last_regex);
  if (error) bad_prog ((char *)error);
}

/* Store a label (or label reference) created by a ':', 'b', or 't'
   comand so that the jump to/from the lable can be backpatched after
   compilation is complete */
sed_label *Engine::setup_jump (struct sed_label *list, struct sed_cmd *cmd, struct vector *vec)
{
  struct sed_label *tmp;
  VOID *b;
  int ch;

  b = init_buffer ();
  while ((ch = inchar ()) != EOF && isblank (ch))
    ;
  /* Possible non posixicity. */
  while (ch != EOF && ch != '\n' && (!isblank (ch)) && ch != ';' && ch != '}')
    {
      add1_buffer (b, ch);
      ch = inchar ();
    }
  savchar (ch);
  add1_buffer (b, '\0');
  tmp = (struct sed_label *) ck_malloc (sizeof (struct sed_label));
  tmp->v = vec;
  tmp->v_index = cmd - vec->v;
  tmp->name = ck_strdup (get_buffer (b));
  tmp->next = list;
  flush_buffer (b);
  return tmp;
}

/* read in a filename for a 'r', 'w', or 's///w' command, and
   update the internal structure about files.  The file is
   opened if it isn't already open. */
FILE *Engine::compile_filename (int readit)
{
  char *file_name;
  int n;
  VOID *b;
  int ch;

  if (inchar () != ' ')
    bad_prog ("missing ' ' before filename");
  b = init_buffer ();
  while ((ch = inchar ()) != EOF && ch != '\n')
    add1_buffer (b, ch);
  add1_buffer (b, '\0');
  file_name = get_buffer (b);
  for (n = 0; n < NUM_FPS; n++)
    {
      if (!file_ptrs[n].name)
   break;
    }
  if (n < NUM_FPS)
    {
      file_ptrs[n].name = ck_strdup (file_name);
      if (!readit)
   {
     if (!file_ptrs[n].for_write)
       file_ptrs[n].for_write = ck_fopen (file_name, "w");
   }
      else
   {
     if (!file_ptrs[n].for_read)
       file_ptrs[n].for_read = fopen (file_name, "r");
   }
      flush_buffer (b);
      return readit ? file_ptrs[n].for_read : file_ptrs[n].for_write;
    }
  else
    {
      bad_prog ("Hopelessely evil compiled in limit on number of open files.  re-compile sed");
      return 0;
    }
}

void Engine::ausgabe(char *ptr,int len)
{
   if (output_art && ptr && len) 
   {
      int lenalt=output_cstr?strlen(output_cstr):0;
      char *p=(char*)malloc(lenalt+len+1);
      if (lenalt) strcpy(p, output_cstr);
      strncpy(p+lenalt, ptr, len); p[lenalt+len]=0;
      free(output_cstr);
      output_cstr=p;
   }
}

void Engine::read_string(char *name, int flagcnt, int flagperline)
{
   if (!name) return;
   input_puffer=name;
   if (flagcnt) { m_line=0; m_globalPosition=0; }
   int len;
   while (read_pattern_space(flagperline))
   {
      if (flagcnt) { m_line++; len=line.length; }
      
      execute_program(the_program);
      
      if (!no_default_output) ausgabe(line.text, line.length);
      if (append.length) { ck_fwrite (append.text, 1, append.length, stdout); append.length = 0; }

      if (quit_cmd) break;

      if (flagcnt)
      {
         if (!m_noMatch) { m_globalPosition+=m_position; break; }
         m_globalPosition+=len;
      }
   } 
}

static char *eol_pos (char *str, int len)
{
   while (len--) if (*str++ == '\n') return --str;
   return --str;
}

static void chr_copy (char *dest, char *src, int len)
{
   while (len--) *dest++ = *src++;
}

void Engine::execute_program (struct vector *vec)
{
   struct sed_cmd *cur_cmd;
   int n;
   int addr_matched;
   static int end_cycle;
   int start, remain, offset;

   static struct line tmp;
   struct line t;
   char *rep, *rep_end, *rep_next, *rep_cur;

   int count;
   struct vector *restart_vec = vec;

restart:
   vec = restart_vec;
   count=0;
   end_cycle = 0;

   for (cur_cmd = vec->v, n = vec->v_length; n; cur_cmd++, n--)
   {
exe_loop:
      addr_matched = 0;
      if (cur_cmd->aflags & A1_MATCHED_BIT)
      {
         addr_matched = 1;
         if (match_address (&(cur_cmd->a2))) cur_cmd->aflags &= ~A1_MATCHED_BIT;
      }
      else if (match_address (&(cur_cmd->a1)))
      {
         addr_matched = 1;
         if (cur_cmd->a2.addr_type != addr_is_null)
            if ((cur_cmd->a2.addr_type == addr_is_regex) || !match_address (&(cur_cmd->a2)))
               cur_cmd->aflags |= A1_MATCHED_BIT;
      }
      if (cur_cmd->aflags & ADDR_BANG_BIT) addr_matched = !addr_matched;
      if (!addr_matched) continue;
      
      switch (cur_cmd->cmd)
      {
      case '{':      /* Execute sub-program */
        if (cur_cmd->x.sub->v_length)
        {
           vec = cur_cmd->x.sub;
           cur_cmd = vec->v;
           n = vec->v_length;
           goto exe_loop;
        }
        break;
   
      case '}':
        cur_cmd = vec->return_v->v + vec->return_i;
        n = vec->return_v->v_length - vec->return_i;
        vec = vec->return_v;
        break;
   
      case ':':      /* Executing labels is easy. */
        break;
   
      case '=':
        fprintf (stdout, "%d\n", input_line_number);
        break;
   
      case 'a':
        while (append.alloc - append.length < cur_cmd->x.cmd_txt.text_len)
        {
           append.alloc *= 2;
           append.text = (char*)ck_realloc (append.text, append.alloc);
        }
        bcopy (cur_cmd->x.cmd_txt.text, append.text + append.length, cur_cmd->x.cmd_txt.text_len);
        append.length += cur_cmd->x.cmd_txt.text_len;
        break;
   
      case 'b':
        if (!cur_cmd->x.jump)
           end_cycle++;
        else
        {
           struct sed_label *j = cur_cmd->x.jump;
   
           n = j->v->v_length - j->v_index;
           cur_cmd = j->v->v + j->v_index;
           vec = j->v;
           goto exe_loop;
        }
        break;
   
      case 'c':
        line.length = 0;
        if (!((cur_cmd->aflags & A1_MATCHED_BIT)))
           ck_fwrite (cur_cmd->x.cmd_txt.text, 1, cur_cmd->x.cmd_txt.text_len, stdout);
        end_cycle++;
        break;
   
      case 'd':
        line.length = 0;
        end_cycle++;
        break;
   
      case 'D': {
        char *tmp;
        int newlength;
   
        tmp = eol_pos (line.text, line.length);
        newlength = line.length - (tmp - line.text) - 1;
        if (newlength)
        {
           chr_copy (line.text, tmp + 1, newlength);
           line.length = newlength;
           goto restart;
        }
        line.length = 0;
        end_cycle++;
        } break;
   
      case 'g':
        line_copy (&hold, &line);
        break;
   
      case 'G':
        line_append (&hold, &line);
        break;
   
      case 'h':
        line_copy (&line, &hold);
        break;
   
      case 'H':
        line_append (&line, &hold);
        break;
   
      case 'i':
        ck_fwrite (cur_cmd->x.cmd_txt.text, 1, cur_cmd->x.cmd_txt.text_len, stdout);
        break;
   
      case 'l': {
        char *tmp;
        int n;
        int width = 0;
   
        n = line.length;
        tmp = line.text;
        while (n--)
        {
           /* Skip the trailing newline, if there is one */
           if (!n && (*tmp == '\n')) break;
           if (width > 77)
           {
              width = 0;
              putc ('\n', stdout);
           }
           if (*tmp == '\\')
           {
              fprintf (stdout,"\\\\");
              width += 2;
           }
           else if (isprint (*tmp))
           {
              putc (*tmp, stdout);
              width++;
           }
           else
              switch (*tmp)
              {
              case 007:  fprintf (stdout, "\\a"); width += 2; break;
              case '\b': fprintf (stdout, "\\b"); width += 2; break;
              case '\f': fprintf (stdout, "\\f"); width += 2; break;
              case '\n': fprintf (stdout, "\\n"); width += 2; break;
              case '\r': fprintf (stdout, "\\r"); width += 2; break;
              case '\t': fprintf (stdout, "\\t"); width += 2; break;
              case '\v': fprintf (stdout, "\\v"); width += 2; break;
              default:   fprintf (stdout, "\\%02x", (*tmp) & 0xFF); width += 2; break;
              }
           tmp++;
        }
        putc ('\n', stdout);
        } break;
   
      case 'n':
        if (myeof()) goto quit;
        if (!no_default_output) ausgabe(line.text, line.length);
        read_pattern_space ();
        break;
   
      case 'N':
        if (myeof()) { line.length = 0; goto quit; }
        append_pattern_space ();
        break;
   
      case 'p':
        ck_fwrite (line.text, 1, line.length, stdout);
        break;
   
      case 'P': {
        char *tmp;
        tmp = eol_pos (line.text, line.length);
        ck_fwrite (line.text, 1, tmp ? tmp - line.text + 1 : line.length, stdout);
        } break;
   
      case 'q':
quit:   quit_cmd++;
        end_cycle++;
        break;
   
      case 'r': {
        int n = 0;
        if (cur_cmd->x.io_file)
        {
           rewind (cur_cmd->x.io_file);
           do
           {
              append.length += n;
              if (append.length == append.alloc)
              {
                 append.alloc *= 2;
                 append.text = (char*)ck_realloc (append.text, append.alloc);
              }
              n = fread (append.text + append.length, sizeof (char), append.alloc - append.length, cur_cmd->x.io_file);
           }
           while (n > 0);
              if (ferror (cur_cmd->x.io_file)) panic ("Read error on input file to 'r' command");
        }
        } break;
   
      case 's': {
        int trail_nl_p = line.text [line.length - 1] == '\n';
        if (!tmp.alloc) { tmp.alloc = 50; tmp.text = (char*)ck_malloc (50); }
        start = 0;
        count = 0;
        remain = line.length - trail_nl_p;
        tmp.length = 0;
        rep = cur_cmd->x.cmd_regex.replacement;
        rep_end = rep + cur_cmd->x.cmd_regex.replace_length;
   
        while ((offset=re_search (cur_cmd->x.cmd_regex.regx, line.text, line.length - trail_nl_p, start, remain, &regs)) >= 0)
        {
           count++;
           if (offset - start) str_append (&tmp, line.text + start, offset - start);
           if (cur_cmd->x.cmd_regex.flags & S_NUM_BIT)
              if (count != cur_cmd->x.cmd_regex.numb)
              {
                 int matched = regs.end[0] - regs.start[0];
                 if (!matched) matched = 1;
                 str_append (&tmp, line.text + regs.start[0], matched);
                 start = (offset == regs.end[0] ? offset + 1 : regs.end[0]);
                 remain = (line.length - trail_nl_p) - start;
                 continue;
              }
           for (rep_next = rep_cur = rep; rep_next < rep_end; rep_next++)
           {
              /*if (*rep_next=='&')
              {
                 if (rep_next - rep_cur) str_append (&tmp, rep_cur, rep_next - rep_cur);
                 str_append (&tmp, line.text + regs.start[0], regs.end[0] - regs.start[0]);
                 rep_cur = rep_next + 1;
              }
              else */if (*rep_next == '\\')
              {
                 if (rep_next - rep_cur) str_append (&tmp, rep_cur, rep_next - rep_cur);
                 rep_next++;
                 if (rep_next != rep_end)
                 {
                    int n;
                    if (*rep_next >= '0' && *rep_next <= '9')
                    {
                       n = *rep_next - '0';
                       str_append (&tmp, line.text + regs.start[n], regs.end[n] - regs.start[n]);
                    }
                    else
                       str_append (&tmp, rep_next, 1);
                 }
                 rep_cur = rep_next + 1;
              }
           }
           if (rep_next - rep_cur) str_append (&tmp, rep_cur, rep_next - rep_cur);
           if (offset == regs.end[0])
           {
              str_append (&tmp, line.text + offset, 1);
              ++regs.end[0];
           }
           start = regs.end[0];
           remain = (line.length - trail_nl_p) - start;
           if (remain < 0) break;
           if (!(cur_cmd->x.cmd_regex.flags & S_GLOBAL_BIT)) break;
        }
        if (!count) break;
        replaced = 1;
        str_append (&tmp, line.text + start, remain + trail_nl_p);
        t.text = line.text;
        t.length = line.length;
        t.alloc = line.alloc;
        line.text = tmp.text;
        line.length = tmp.length;
        line.alloc = tmp.alloc;
        tmp.text = t.text;
        tmp.length = t.length;
        tmp.alloc = t.alloc;
        if ((cur_cmd->x.cmd_regex.flags & S_WRITE_BIT) && cur_cmd->x.cmd_regex.wio_file)
           ck_fwrite (line.text, 1, line.length, cur_cmd->x.cmd_regex.wio_file);
        if (cur_cmd->x.cmd_regex.flags & S_PRINT_BIT)
           ck_fwrite (line.text, 1, line.length, stdout);
        } break;
          
      case 'f':
      case 'z': 
      {
        int trail_nl_p = line.text [line.length - 1] == '\n';
        if (!tmp.alloc) { tmp.alloc = 50; tmp.text = (char*)ck_malloc (50); }
        count = 0;
        start = 0;
        remain = line.length - trail_nl_p;
        tmp.length = 0;
        rep = cur_cmd->x.cmd_regex.replacement;
        rep_end = rep + cur_cmd->x.cmd_regex.replace_length;
   
        int s0=start;
        while ((offset=re_search (cur_cmd->x.cmd_regex.regx, line.text, line.length - trail_nl_p, start, remain, &regs)) >= 0)
        {
           count++;
           if (cur_cmd->x.cmd_regex.flags & S_NUM_BIT)
           {
              if (count != cur_cmd->x.cmd_regex.numb)
              {
                 int matched = regs.end[0] - regs.start[0];
                 if (!matched) matched = 1;
                 start = (offset == regs.end[0] ? offset + 1 : regs.end[0]);
                 remain = (line.length - trail_nl_p) - start;
                 continue;
              }
           }
           if (cur_cmd->cmd=='z')
              output_count++;
           else
              if (!(--findnr_tmp)) { m_noMatch=FALSE; m_position=offset-s0+1; break; }

           for (rep_next=rep_cur=rep; rep_next<rep_end; rep_next++)
           {
              /*if (*rep_next=='&')       rep_cur=rep_next+1;
              else */if (*rep_next=='\\') rep_cur=(++rep_next)+1;
           }
           if (offset==regs.end[0]) ++regs.end[0];
           start=regs.end[0];
           remain=(line.length-trail_nl_p)-start;
           if (remain<0) break;
           if (!(cur_cmd->x.cmd_regex.flags & S_GLOBAL_BIT)) break;
        }
        replaced = 1;
        if (cur_cmd->cmd!='z' && !m_noMatch) break;
        t.text = line.text;
        t.length = line.length;
        t.alloc = line.alloc;
        line.text = tmp.text;
        line.length = tmp.length;
        line.alloc = tmp.alloc;
        tmp.text = t.text;
        tmp.length = t.length;
        tmp.alloc = t.alloc;
      } 
        break;
          
      case 't':
         if (replaced)
         {
            replaced = 0;
            if (!cur_cmd->x.jump) 
               end_cycle++;
            else
            {
               struct sed_label *j = cur_cmd->x.jump;
               n = j->v->v_length - j->v_index;
               cur_cmd = j->v->v + j->v_index;
               vec = j->v;
               goto exe_loop;
            }
         }
         break;
   
      case 'w':
        if (cur_cmd->x.io_file)
        {
           ck_fwrite (line.text, 1, line.length, cur_cmd->x.io_file);
           fflush (cur_cmd->x.io_file);
        }
        break;
   
      case 'x': {
        struct line tmp;
        tmp = line;
        line = hold;
        hold = tmp;
        } break;
   
      case 'y': {
        unsigned char *p, *e;
        for (p = (unsigned char *) (line.text), e = p + line.length; p < e; p++)
           *p = cur_cmd->x.translate[*p];
        } break;
   
      default:
        panic ("INTERNAL ERROR: Bad cmd %c", cur_cmd->cmd);
      }
      
      if (end_cycle) break;
   }
}

/* Return non-zero if the current line matches the address pointed to by 'addr'. */
int Engine::match_address (struct addr *addr)
{
  switch (addr->addr_type)
    {
    case addr_is_null:
      return 1;
    case addr_is_num:
      return (input_line_number == addr->addr_number);
    case addr_is_mod:
      return ((input_line_number%addr->modulo) == addr->offset);


    case addr_is_regex:
      {
   int trail_nl_p = line.text [line.length - 1] == '\n';
   return (re_search (addr->addr_regex,
            line.text,
            line.length - trail_nl_p,
            0,
            line.length - trail_nl_p,
            (struct re_registers *) 0) >= 0) ? 1 : 0;
      }
    case addr_is_last:
      return (input_EOF) ? 1 : 0;

    default:
      panic ("INTERNAL ERROR: bad address type");
      break;
    }
  return -1;
}

/* Read in the next line of input, and store it in the pattern space.  Return non-zero if this is the last line of input */
int Engine::read_pattern_space (int flagperline)
{
   char *p = line.text;
   int n = line.alloc;
   int ch;

   if (myeof()) return 0;
   input_line_number++;
   replaced=0;
   for (;;)
   {
      if (n==0)
      {
         line.text=(char*)ck_realloc (line.text, line.alloc * 2);
         p=line.text+line.alloc;
         n=line.alloc;
         line.alloc*=2;
      }
      ch = mygetc();
      if (mytesteof(ch))
      {
         if (n == line.alloc) return 0;
         /* *p++ = '\n'; */
         /* --n; */
         line.length = line.alloc - n;
         if (last_input_file) input_EOF++;
         return 1;
      }
      // if (ch!='\r') { 
      *p++=ch; --n; 
      // }
      if (ch=='\n' && flagperline)
      {
         line.length = line.alloc - n;
         break;
      }
   }
   ch=mygetc();
   if (!mytesteof(ch)) myungetc(ch);
   else if (last_input_file) input_EOF++;
   return 1;
}

/* Inplement the 'N' command, which appends the next line of input to
   the pattern space. */
void Engine::append_pattern_space ()
{
   char *p;
   int n;
   int ch;
   p = line.text + line.length;
   n = line.alloc - line.length;
   input_line_number++;
   replaced = 0;
   for (;;)
   {
      ch = mygetc();
      if (mytesteof(ch))
      {
         if (n == line.alloc) return;
         /* *p++ = '\n'; */
         /* --n; */
         line.length = line.alloc - n;
         if (last_input_file) input_EOF++;
         return;
      }
      if (n == 0)
      {
         line.text = (char*)ck_realloc (line.text, line.alloc * 2);
         p = line.text + line.alloc;
         n = line.alloc;
         line.alloc *= 2;
      }
      *p++ = ch;
      --n;
      if (ch == '\n')
      {
         line.length = line.alloc - n;
         break;
      }
   }
   ch = mygetc();
   if (!mytesteof(ch)) 
      ungetc (ch, input_file);
   else if (last_input_file) 
      input_EOF++;
}

/* Copy the contents of the line 'from' into the line 'to'.
   This destroys the old contents of 'to'.  It will still work
   if the line 'from' contains nulls. */
void line_copy (struct line *from, struct line *to)
{
  if (from->length > to->alloc)
    {
      to->alloc = from->length;
      to->text = (char*)ck_realloc (to->text, to->alloc);
    }
  bcopy (from->text, to->text, from->length);
  to->length = from->length;
}

/* Append the contents of the line 'from' to the line 'to'.
   This routine will work even if the line 'from' contains nulls */
void line_append (struct line *from, struct line *to)
{
  if (from->length > (to->alloc - to->length))
    {
      to->alloc += from->length;
      to->text = (char*)ck_realloc (to->text, to->alloc);
    }
  bcopy (from->text, to->text + to->length, from->length);
  to->length += from->length;
}

/* Append 'length' bytes from 'string' to the line 'to'
   This routine *will* append bytes with nulls in them, without failing. */
void str_append (struct line *to, char *string, int length)
{
  if (length > to->alloc - to->length)
    {
      to->alloc += length;
      to->text = (char*)ck_realloc (to->text, to->alloc);
    }
  if (length>0) bcopy (string, to->text + to->length, length);
  to->length += length;
}

void Engine::init(void)
{
   input_line_number = 0;
   last_input_file = 0;
   input_EOF = 0;
   quit_cmd = 0;
   replaced = 0;
   program_depth = 0;
   the_program = 0;
   jumps = 0;
   labels = 0;
   prog_line = 1;
   bad_input = 0;

   no_default_output = 0; /* If set, don't write out the line unless explictly told to */
   output_cstr=0;
   output_count=0;
   regs.start=0; 
   regs.num_regs=0; 
   regs.end=0;
   last_regex_string=0;

   for (int nf = 0; nf < NUM_FPS; nf++)
   {
      file_ptrs[nf].for_write=0;
      file_ptrs[nf].for_read=0;
   }
}

int Engine::myeof(void) { return (!*input_puffer); }

int Engine::mygetc(void) { return *(input_puffer++); }

int Engine::mytesteof(int c) { return c==0; }

void Engine::myungetc(int ch) { input_puffer--; }

void Engine::close_files (void)
{
   for (int nf = 0; nf < NUM_FPS; nf++)
   {
      if (file_ptrs[nf].for_write) fclose (file_ptrs[nf].for_write);
      if (file_ptrs[nf].for_read) fclose (file_ptrs[nf].for_read);
   }
}

